import java.util.*;
import java.io.*;
import localsolver.*;

public class CapacitatedFacilityLocation {

    // Data from the problem
    private int nbMaxFacilities;
    private int nbSites;
    private double[] capacityData;
    private double[] openingPriceData;
    private double[] demandData;
    private double[][] allocationPriceData;

    // LocalSolver
    private final LocalSolver localsolver;

    // Variables
    private LSExpression[] facilityAssignments;

    // Objective
    private LSExpression[] cost;
    private LSExpression totalCost;

    private CapacitatedFacilityLocation(LocalSolver localsolver) {
        this.localsolver = localsolver;
    }

    /* Read instance data */
    private void readInstance(String fileName) throws IOException {
        try (Scanner input = new Scanner(new File(fileName))) {
            nbMaxFacilities = input.nextInt();
            nbSites = input.nextInt();
            capacityData = new double[nbMaxFacilities];
            openingPriceData = new double[nbMaxFacilities];
            demandData = new double[nbSites];
            allocationPriceData = new double[nbMaxFacilities][nbSites];

            for (int f = 0; f < nbMaxFacilities; ++f) {
                // List of facilities capacities
                capacityData[f] = input.nextDouble();
                // List of fixed costs induced by the facilities opening
                openingPriceData[f] = input.nextDouble();
            }

            // Demand of each site
            for (int s = 0; s < nbSites; ++s) {
                demandData[s] = input.nextDouble();
            }

            // Allocation price between sites and facilities
            for (int f = 0; f < nbMaxFacilities; ++f) {
                for (int s = 0; s < nbSites; ++s) {
                    allocationPriceData[f][s] = input.nextDouble();
                }
            }
        }
    }

    private void solve(int limit) {
        // Declare the optimization model
        LSModel model = localsolver.getModel();

        // Facilities are represented by the sets of sites they provide
        facilityAssignments = new LSExpression[nbMaxFacilities];
        for (int f = 0; f < nbMaxFacilities; ++f) {
            facilityAssignments[f] = model.setVar(nbSites);
        }

        // Each site is covered by exactly one facility
        model.constraint(model.partition(facilityAssignments));

        // Converting demand and allocationPrice into LocalSolver array
        LSExpression demand = model.array(demandData);
        LSExpression allocationPrice = model.array(allocationPriceData);

        cost = new LSExpression[nbMaxFacilities];
        for (int f = 0; f < nbMaxFacilities; ++f) {
            LSExpression fExpr = model.createConstant(f);

            LSExpression facility = facilityAssignments[f];
            LSExpression size = model.count(facility);

            // Capacity constraint
            LSExpression demandLambda = model.lambdaFunction(i -> model.at(demand, i));
            model.constraint(model.leq(model.sum(facility, demandLambda), capacityData[f]));

            // Cost (allocation price + opening price)
            LSExpression costLambda = model.lambdaFunction(i -> model.at(allocationPrice, fExpr, i));
            cost[f] = model.sum(model.sum(facility, costLambda), model.iif(model.gt(size, 0), openingPriceData[f], 0));
        }

        // Objective : minimize total cost
        totalCost = model.sum(cost);
        model.minimize(totalCost);

        model.close();

        // Parameterize the solver
        localsolver.getParam().setTimeLimit(limit);

        localsolver.solve();
    }

    /*
     * Write the solution in a file with the following format:
     * - value of the objective
     * - indices of the open facilities followed by all the sites they provide
     */
    private void writeSolution(String file_name) throws IOException {
        try (PrintWriter output = new PrintWriter(file_name)) {
            output.println(totalCost.getDoubleValue());
            for (int f = 0; f < nbMaxFacilities; ++f) {
                if (cost[f].getDoubleValue() > 0) {
                    output.println(f);
                    LSCollection sites_assigned = facilityAssignments[f].getCollectionValue();
                    for (int s = 0; s < sites_assigned.count(); ++s) {
                        output.print(sites_assigned.get(s) + " ");
                    }
                    output.println();
                }
            }
        }
    }

    public static void main(String[] args) {
        try (LocalSolver localsolver = new LocalSolver()) {
            String instanceFile = args[0];
            String outputFile = args.length > 1 ? args[1] : null;
            String strTimeLimit = args.length > 2 ? args[2] : "20";

            CapacitatedFacilityLocation model = new CapacitatedFacilityLocation(localsolver);
            model.readInstance(instanceFile);
            model.solve(Integer.parseInt(strTimeLimit));
            if (outputFile != null) {
                model.writeSolution(outputFile);
            }
        } catch (Exception ex) {
            System.err.println(ex);
            ex.printStackTrace();
            System.exit(1);
        }
    }
}