use io;

/* Read instance data */
function input() {
    usage = "Usage: localsolver capacitated_facility_location.lsp "
            + "inFileName=inputFile [solFileName=outputFile] [lsTimeLimit=timeLimit]";

    if (inFileName == nil) throw usage;
    local inFile = io.openRead(inFileName);
    nbMaxFacilities = inFile.readInt();
    nbSites = inFile.readInt();

    for [f in 0...nbMaxFacilities] {
        // List of facilities capacities
        capacity[f] = inFile.readDouble();
        // List of fixed costs induced by the facilities opening
        openingPrice[f] = inFile.readDouble();
    }

    // Demand of each site
    demand[s in 0...nbSites] = inFile.readDouble();

    // Allocation price between sites and facilities
    allocationPrice[f in 0...nbMaxFacilities][s in 0...nbSites] = inFile.readDouble();
}

/* Declare the optimization model */
function model() {
    // Facilities are represented by the set of sites they provide
    facilityAssignments[f in 0...nbMaxFacilities] <- set(nbSites);

    // Each site is covered by exactly one facility
    constraint partition[f in 0...nbMaxFacilities](facilityAssignments[f]);

    for [f in 0...nbMaxFacilities] {
        local facility <- facilityAssignments[f];
        local size <- count(facility);

        // Capacity constraint
        constraint sum(facility, i => demand[i]) <= capacity[f];

        // Cost (allocation price + opening price)
        cost[f] <- sum(facility, i => allocationPrice[f][i]) + openingPrice[f] * (size > 0);
    }

    // Objective : minimize total cost
    totalCost <- sum[f in 0...nbMaxFacilities](cost[f]);
    minimize totalCost;
}

/* Parametrize the solver */
function param() {
    if (lsTimeLimit == nil) lsTimeLimit = 20;
}

/* Write the solution in a file with the following format:
 * - value of the objective
 * - indices of the open facilities followed by all the sites they provide */
function output() {
    if (solFileName == nil) return;
    local outfile = io.openWrite(solFileName);
    outfile.println(totalCost.value);
    for [f in 0...nbMaxFacilities : cost[f].value > 0] {
        outfile.println(f);
        for [s in facilityAssignments[f].value]
            outfile.print(s, " ");
        outfile.println();
    }
}